/***************************************************************************
 $RCSfile: mediumddv2.cpp,v $
                             -------------------
    cvs         : $Id: mediumddv2.cpp,v 1.1.1.1 2004/06/04 11:08:21 aquamaniac Exp $
    begin       : Fri Nov 16 2001
    copyright   : (C) 2001 by Martin Preuss
    email       : openhbci@aquamaniac.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

/*
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef __declspec
# if BUILDING_DLL
#  define DLLIMPORT __declspec (dllexport)
# else /* Not BUILDING_DLL */
#  define DLLIMPORT __declspec (dllimport)
# endif /* Not BUILDING_DLL */
#else
# define DLLIMPORT
#endif



#include <openssl/bn.h>
#include <openssl/ripemd.h>


#include <list>
#include <openhbci2/error.h>
#include <openhbci2/pointer.h>
#include <openhbci2/hbcistring.h>
#include <openhbci2/date.h>
#include <openhbci2/value.h>
#include <openhbci2/auth.h>
#include <openhbci2/interactor.h>
#include <openhbci2/hbci.h>
#include <openhbci2/user.h>
#include <openhbci2/bank.h>
#include <openhbci2/cryptkey.h>
#include <openhbci2/deskey.h>
#include <openhbci2/medium.h>
#include <mediumddv2.h>

#include <unistd.h> // for sleep


// for bad_cast exception
#include <typeinfo>

namespace HBCI {


/*_________________________________________________________________________
 *AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
 *                              MediumDDV2
 *YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
 */



MediumDDV2::MediumDDV2(DDVCard2Plugin *plugin,
		       const Hbci *hbci,
		       const string &cardnumber)
:Medium(hbci)
,_plugin(plugin)
,_card(0)
,_ismounted(false)
,_mountcount(0)
,_useKeyPad(true)
,_cardnumber(cardnumber)
,_country(280)
,_signKeyNumber(2)
,_signKeyVersion(1)
,_cryptKeyNumber(3)
,_cryptKeyVersion(1)
{
}


MediumDDV2::~MediumDDV2(){
}


const string& MediumDDV2::mediumName() const {
  return _cardnumber;
}


const string& MediumDDV2::serialNumber() const {
  return _cardnumber;
}


Error MediumDDV2::verify(const string &data, const string &signature){
  GWEN_BUFFER *mbuf;
  GWEN_BUFFER *shbuf;
  LC_CLIENT_RESULT res;
  string mac;
  string hash;

  // make RIPEMD-160
  hash=ripe(data);
  mbuf=GWEN_Buffer_new(0, hash.length(), 0, 1);
  shbuf=GWEN_Buffer_new(0, 8, 0, 1);

  GWEN_Buffer_AppendBytes(mbuf, hash.data(), hash.length());
  res=LC_DDVCard_SignHash(_card, mbuf, shbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(0, "Error signing hash (%d)", res);
    GWEN_Buffer_free(shbuf);
    GWEN_Buffer_free(mbuf);
    return Error("MediumDDV2::verify",
		 "Error signing hash",
		 0);
  }

  mac=string(GWEN_Buffer_GetStart(shbuf), GWEN_Buffer_GetUsedBytes(shbuf));
  GWEN_Buffer_free(shbuf);
  GWEN_Buffer_free(mbuf);

  // compare both strings
  if (mac==signature) {
    return Error();
  }
  return Error("MediumDDV2::verify",
	       "Bad signature()",
	       0);
}


string MediumDDV2::sign(const string &data){
  GWEN_BUFFER *mbuf;
  GWEN_BUFFER *shbuf;
  LC_CLIENT_RESULT res;
  string mac;
  string hash;

  // make RIPEMD-160
  hash=ripe(data);
  mbuf=GWEN_Buffer_new(0, hash.length(), 0, 1);
  shbuf=GWEN_Buffer_new(0, 8, 0, 1);

  GWEN_Buffer_AppendBytes(mbuf, hash.data(), hash.length());
  res=LC_DDVCard_SignHash(_card, mbuf, shbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(0, "Error signing hash (%d)", res);
    GWEN_Buffer_free(shbuf);
    GWEN_Buffer_free(mbuf);
    throw Error("MediumDDV2::verify",
		"Error signing hash",
		0);
  }

  mac=string(GWEN_Buffer_GetStart(shbuf), GWEN_Buffer_GetUsedBytes(shbuf));
  GWEN_Buffer_free(shbuf);
  GWEN_Buffer_free(mbuf);

  return mac;
}


string MediumDDV2::createMessageKey() const {
  GWEN_BUFFER *kbuf;
  LC_CLIENT_RESULT res;
  string key;

  kbuf=GWEN_Buffer_new(0, 16, 0, 1);
  // create left
  res=LC_DDVCard_GetChallenge(_card, kbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(0, "Error getting random-1 (%d)", res);
    GWEN_Buffer_free(kbuf);
    throw Error("MediumDDV2::createMessageKey",
		"Error on LC_DDVCard_GetChallenge()",
		0);
  }

  // create right
  res=LC_DDVCard_GetChallenge(_card, kbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(0, "Error getting random-2 (%d)", res);
    GWEN_Buffer_free(kbuf);
    throw Error("MediumDDV2::createMessageKey",
		"Error on LC_DDVCard_GetChallenge()",
		0);
  }

  key=string(GWEN_Buffer_GetStart(kbuf), GWEN_Buffer_GetUsedBytes(kbuf));
  GWEN_Buffer_free(kbuf);
  return key;
}


string MediumDDV2::encryptKey(const string &srckey){
  LC_CLIENT_RESULT res;
  GWEN_BUFFER *ibuf;
  GWEN_BUFFER *obuf;
  string kL;
  string kR;
  string kLenc;
  string kRenc;
  string dstkey;

  // check sizes
  if (srckey.length()!=16)
    throw Error("MediumDDV2::encryptKey",
		"Bad length of srckey",
		0);
  // split key
  kL=srckey.substr(0,8);
  kR=srckey.substr(8);
  ibuf=GWEN_Buffer_new(0, 8, 0, 1);
  obuf=GWEN_Buffer_new(0, 8, 0, 1);

  // encode left
  GWEN_Buffer_AppendBytes(ibuf, kL.data(), kL.length());
  res=LC_DDVCard_CryptBlock(_card, ibuf, obuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(0, "Error encrypting block (%d)", res);
    GWEN_Buffer_free(obuf);
    GWEN_Buffer_free(ibuf);
    throw Error("MediumDDV2::encryptKey",
		"Error on LC_DDVCard_CryptBlock()",
		0);
  }
  dstkey=string(GWEN_Buffer_GetStart(obuf), GWEN_Buffer_GetUsedBytes(obuf));

  // encode right
  GWEN_Buffer_Reset(ibuf);
  GWEN_Buffer_Reset(obuf);
  GWEN_Buffer_AppendBytes(ibuf, kR.data(), kR.length());
  res=LC_DDVCard_CryptBlock(_card, ibuf, obuf);
  if (res!=LC_Client_ResultOk) {
    DBG_ERROR(0, "Error encrypting block (%d)", res);
    GWEN_Buffer_free(obuf);
    GWEN_Buffer_free(ibuf);
    throw Error("MediumDDV2::encryptKey",
		"Error on LC_DDVCard_CryptBlock()",
		0);
  }
  dstkey+=string(GWEN_Buffer_GetStart(obuf), GWEN_Buffer_GetUsedBytes(obuf));
  return dstkey;
}



string MediumDDV2::decryptKey(const string &srckey){
  return encryptKey(srckey);
}


Error MediumDDV2::_enterPin(LC_CARD *hcard,
			   const string &pin) {
  Error herr;
  string localpin;
  LC_CLIENT_RESULT res;

  if ((LC_Card_GetReaderFlags(hcard) & LC_CARD_READERFLAGS_KEYPAD) &&
      _useKeyPad && pin.empty()) {
    if (Hbci::debugLevel()>5)
      fprintf(stderr,"MediumDDV2: Terminal has a keypad, will ask for pin.\n");
    // tell the user about pin verification
    hbci()->interactor().ref().
      msgStartInputPinViaKeypad(owner());

    res=LC_DDVCard_SecureVerifyPin(hcard);
    // tell the user about end of pin verification
    hbci()->interactor().ref().
      msgFinishedInputPinViaKeypad(owner());
    if (res!=LC_Client_ResultOk) {
      string errString;
      const char *p;
      char numbuf[16];
      int ec;

      if (Hbci::debugLevel()>5)
	fprintf(stderr,"MediumDDV2: Bad pin.\n");

      errString="Error: SW1=";
      snprintf(numbuf, sizeof(numbuf), "%02x", LC_Card_GetLastSW1(hcard));
      errString+=numbuf;
      errString+=" SW2=";
      snprintf(numbuf, sizeof(numbuf), "%02x", LC_Card_GetLastSW2(hcard));
      errString+=numbuf;
      p=LC_Card_GetLastText(hcard);
      if (p) {
	errString+=" (";
	errString+=p;
	errString+=")";
      }
      hbci()->interactor().ref().msgStateResponse(errString);
      if (LC_Card_GetLastSW1(hcard)==0x63) {
	switch (LC_Card_GetLastSW2(hcard)) {
	case 0xc0: ec=HBCI_ERROR_CODE_PIN_WRONG_0; break;
	case 0xc1: ec=HBCI_ERROR_CODE_PIN_WRONG_1; break;
	case 0xc2: ec=HBCI_ERROR_CODE_PIN_WRONG_2; break;
	default:   ec=HBCI_ERROR_CODE_PIN_WRONG; break;
	} // switch
	return Error("MediumDDV2::_enterPin",
		     ERROR_LEVEL_CRITICAL,
		     ec,
		     ERROR_ADVISE_RETRY,
		     "Bad pin",
		     "verifyPin (1)");
      }
      else if (LC_Card_GetLastSW1(hcard)==0x69 &&
	       LC_Card_GetLastSW2(hcard)==0x83) {
	ec=HBCI_ERROR_CODE_CARD_DESTROYED;
	return Error("MediumDDV2::_enterPin",
		     ERROR_LEVEL_CRITICAL,
		     ec,
		     ERROR_ADVISE_ABORT,
		     "Card destroyed",
		     "verifyPin (2)");
      }
      else if (LC_Card_GetLastSW1(hcard)==0x64 &&
	       LC_Card_GetLastSW2(hcard)==0x01) {
	ec=HBCI_ERROR_CODE_PIN_ABORTED;
	return Error("MediumDDV2::_enterPin",
		     ERROR_LEVEL_NORMAL,
		     ec,
		     ERROR_ADVISE_ABORT,
		     "User aborted",
		     "verifyPin (3)");

      }
      else {
	return Error("MediumDDV2::_enterPin",
		     ERROR_LEVEL_NORMAL,
		     HBCI_ERROR_CODE_PIN_ABORTED,
		     ERROR_ADVISE_ABORT,
		     "Unknown error"
		     "verifyPin (4)");
      }
    } // if not ok
  } // if hasKeyPad
  else {
    localpin=pin;
    if (Hbci::debugLevel()>5)
      fprintf(stderr,
	      "MediumDDV2: No keypad (or disabled), will ask for pin.\n");
    if (localpin.length()<4) {
      herr=hbci()->authentificator().ref().getSecret(owner(),
						    _cardnumber,
						    localpin);
      if (!herr.isOk()) {
	if (Hbci::debugLevel()>5)
	  fprintf(stderr,"MediumDDV2: No pin entered.\n");
	hbci()->interactor().ref().msgStateResponse("Error: "+
						   herr.errorString());
	return Error("MediumDDV2::_enterPin",
		     ERROR_LEVEL_NORMAL,
		     herr.code(),
		     ERROR_ADVISE_ABORT,
		     "User aborted",
		     "verifyPin (5)");
      } // if user aborted
    } // if pin is not long enough

    // verify the pin
    if (Hbci::debugLevel()>5)
      fprintf(stderr,"MediumDDV2: Verifying cardholder pin.\n");


    res=LC_DDVCard_VerifyPin(hcard, localpin.c_str());
    // tell the user about end of pin verification
    hbci()->interactor().ref().
      msgFinishedInputPinViaKeypad(owner());
    if (res!=LC_Client_ResultOk) {
      string errString;
      const char *p;
      char numbuf[16];
      int ec;

      if (Hbci::debugLevel()>5)
	fprintf(stderr,"MediumDDV2: Bad pin.\n");

      errString="Error: SW1=";
      snprintf(numbuf, sizeof(numbuf), "%02x", LC_Card_GetLastSW1(hcard));
      errString+=numbuf;
      errString+=" SW2=";
      snprintf(numbuf, sizeof(numbuf), "%02x", LC_Card_GetLastSW2(hcard));
      errString+=numbuf;
      p=LC_Card_GetLastText(hcard);
      if (p) {
	errString+=" (";
	errString+=p;
	errString+=")";
      }
      hbci()->interactor().ref().msgStateResponse(errString);
      if (LC_Card_GetLastSW1(hcard)==0x63) {
	hbci()->authentificator().ref().setSecretStatus
	  (owner(),
	   _cardnumber,
	   localpin,
	   HBCI_AUTH_STATUS_BAD);

	switch (LC_Card_GetLastSW2(hcard)) {
	case 0xc0: ec=HBCI_ERROR_CODE_PIN_WRONG_0; break;
	case 0xc1: ec=HBCI_ERROR_CODE_PIN_WRONG_1; break;
	case 0xc2: ec=HBCI_ERROR_CODE_PIN_WRONG_2; break;
	default:   ec=HBCI_ERROR_CODE_PIN_WRONG; break;
	} // switch
	return Error("MediumDDV2::_enterPin",
		     ERROR_LEVEL_CRITICAL,
		     ec,
		     ERROR_ADVISE_RETRY,
		     "Bad pin",
		     "verifyPin (6)");
      }
      else if (LC_Card_GetLastSW1(hcard)==0x69 &&
	       LC_Card_GetLastSW2(hcard)==0x83) {
	hbci()->authentificator().ref().setSecretStatus
	  (owner(),
	   _cardnumber,
	   localpin,
	   HBCI_AUTH_STATUS_BAD);

	ec=HBCI_ERROR_CODE_CARD_DESTROYED;
	return Error("MediumDDV2::_enterPin",
		     ERROR_LEVEL_CRITICAL,
		     ec,
		     ERROR_ADVISE_ABORT,
		     "Card destroyed",
		     "verifyPin (7)");
      }
      else {
	hbci()->authentificator().ref().setSecretStatus
	  (owner(),
	   _cardnumber,
	   localpin,
	   HBCI_AUTH_STATUS_BAD);
	return Error("MediumDDV2::_enterPin",
		     ERROR_LEVEL_NORMAL,
		     HBCI_ERROR_CODE_PIN_ABORTED,
		     ERROR_ADVISE_ABORT,
		     "Unknown error"
		     "verifyPin (8)");
      }
    } // if not ok
    else {
      // PIN is ok, tell the authentificator
      hbci()->authentificator().ref().setSecretStatus(owner(),
						      _cardnumber,
						      localpin,
						      HBCI_AUTH_STATUS_OK);
    }
  } // if no keyPad
  return Error();
}


Error MediumDDV2::mountMedium(const string &pin){
  bool first;
  Error herr;
  bool havepin;
  LC_CARD *hcard;
  LC_CLIENT_RESULT res;
  GWEN_DB_NODE *dbCardData;
  string currCardNumber;

  if (isMounted()) {
    _mountcount++;
    return Error();
  }

  if (_card) {
    DBG_ERROR(0, "We should not have a card here...");
    return Error("MediumDDV2::mountMedium()",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_INVALID,
		 ERROR_ADVISE_ABORT,
		 "INTERNAL: We should not have a card here");
  }
  res=LC_Client_StartWait(_plugin->client(), 0, 0);
  if (res!=LC_Client_ResultOk) {
    fprintf(stderr, "ERROR: Wait timed out.\n");
    return Error("MediumDDV2::mountMedium()",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_NO_CARD,
		 ERROR_ADVISE_ABORT,
		 "could not start waiting for card");
  }

  first=true;
  for (;;) {
    int timeout;

    // determine timeout value
    if (first)
      timeout=3;
    else
      timeout=5;

    hcard=LC_Client_WaitForNextCard(_plugin->client(), timeout);
    if (!hcard) {
      if (!hbci()->interactor().ref().
	  msgInsertMediumOrAbort(owner(), MediumTypeCard)) {
	if (Hbci::debugLevel()>5)
	  fprintf(stderr,"MediumDDV2: User aborted.\n");
	LC_Client_StopWait(_plugin->client());
	return Error("MediumDDV2::mountMedium()",
		     ERROR_LEVEL_NORMAL,
		     HBCI_ERROR_CODE_USER_ABORT,
		     ERROR_ADVISE_ABORT,
		     "No card, user aborted.");
      }
    }
    else {
      // ok, we have a card, now check it
      if (LC_DDVCard_ExtendCard(hcard)) {
        LC_Card_free(hcard);
	LC_Client_StopWait(_plugin->client());
	return Error("MediumDDV2::mountMedium()",
		     ERROR_LEVEL_NORMAL,
		     HBCI_ERROR_CODE_USER_ABORT,
		     ERROR_ADVISE_ABORT,
		     "Not a DDV card");
      }

      res=LC_Card_Open(hcard);
      if (res!=LC_Client_ResultOk) {
	LC_Card_free(hcard);
        DBG_NOTICE(0, "Could not open card, maybe not a DDV card?");
      } // if card not open
      else {
        dbCardData=LC_DDVCard_GetCardDataAsDb(hcard);
	assert(dbCardData);

        currCardNumber=GWEN_DB_GetCharValue(dbCardData,
                                            "cardNumber",
                                            0,
                                            "");
        if (_cardnumber.empty() || _cardnumber==currCardNumber)
          break;

        LC_Card_Close(hcard);
	LC_Card_free(hcard);

	hcard=LC_Client_PeekNextCard(_plugin->client());

	if (!hcard) {
	  if (!hbci()->interactor().ref().
	      msgInsertCorrectMediumOrAbort(owner(),
					    MediumTypeCard)) {
	    LC_Client_StopWait(_plugin->client());
	    return Error("MediumDDV2::mountMedium()",
			 ERROR_LEVEL_NORMAL,
			 HBCI_ERROR_CODE_USER_ABORT,
			 ERROR_ADVISE_ABORT,
			 "user aborted");
	  }
	} // if there is no other card waiting
	else {
	  /* otherwise there already is another card in another reader,
	   * so no need to bother the user. This allows to insert all
	   * cards in all readers and let me choose the card ;-) */
	} // if there is another card waiting
      } // if card open
    } // if there is a card

    first=false;
  } // for

  // ok, now we have the card we wanted to have, now ask for the pin
  LC_Client_StopWait(_plugin->client());
  havepin=false;
  while(!havepin) {
    // TODO: enter pin
    herr=_enterPin(hcard, pin);
    if (!herr.isOk()) {
      LC_Card_Close(hcard);
      LC_Card_free(hcard);
      return Error("mountMedium", herr);
    }
    else
      havepin=true;
  } // while !havepin

  // ok, now we have a card and the PIN is verified, that's it ;-)
  _card=hcard;
  _mountcount++;
  herr=_readKeys();
  if (!herr.isOk()) {
    _card=0;
    _mountcount=0;
    LC_Card_Close(hcard);
    LC_Card_free(hcard);
    return Error("mountMedium", herr);
  }

  if (_cardnumber.empty() && !currCardNumber.empty())
    _cardnumber=currCardNumber;

  DBG_NOTICE(0, "Medium mounted");
  return Error();
}



Error MediumDDV2::unmountMedium(const string&){
  LC_CLIENT_RESULT res;

  if (--_mountcount>0)
    return Error();
  DBG_NOTICE(0, "Unmounting medium");
  _mountcount=0;
  DBG_NOTICE(0, "Closing card");
  res=LC_Card_Close(_card);
  DBG_NOTICE(0, "Deleting card");
  LC_Card_free(_card);
  _card=0;
  if (res!=LC_Client_ResultOk) {
    return Error("MediumDDV2::unmountMedium()",
		 "Could not unmount chip card.",
		 0);
  }
  DBG_NOTICE(0, "Medium unmounted");
  return Error();
}


bool MediumDDV2::isMounted() {
  if (_mountcount<1)
    return false;
  return true;
}


string MediumDDV2::mediumId() {
  GWEN_BUFFER *buf;
  string result;

  buf=LC_DDVCard_GetCardDataAsBuffer(_card);
  assert(buf);
  result=string(GWEN_Buffer_GetStart(buf), GWEN_Buffer_GetUsedBytes(buf));
  return result;
}


unsigned int MediumDDV2::nextSEQ(){
  LC_CLIENT_RESULT res;
  GWEN_BUFFER *mbuf;
  GWEN_DB_NODE *dbRecord;
  int seq;

  res=LC_ProcessorCard_SelectEF(_card, "EF_SEQ");
  if (res!=LC_Client_ResultOk) {
    DBG_INFO(0, "here");
    return 0;
  }

  mbuf=GWEN_Buffer_new(0, 4, 0, 1);
  res=LC_Card_IsoReadRecord(_card, LC_CARD_ISO_FLAGS_RECSEL_GIVEN, 1, mbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_INFO(0, "here");
    GWEN_Buffer_free(mbuf);
    return 0;
  }
  GWEN_Buffer_Rewind(mbuf);
  dbRecord=GWEN_DB_Group_new("seq");
  if (LC_Card_ParseRecord(_card, 1, mbuf, dbRecord)) {
    DBG_ERROR(0, "Error parsing record");
    GWEN_DB_Group_free(dbRecord);
    GWEN_Buffer_free(mbuf);
    return 0;
  }

  seq=GWEN_DB_GetIntValue(dbRecord, "seq", 0, -1);
  if (seq==-1) {
    DBG_ERROR(0, "Bad record data in EF_SEQ");
    GWEN_DB_Group_free(dbRecord);
    GWEN_Buffer_free(mbuf);
    return 0;
  }

  seq++;
  GWEN_DB_SetIntValue(dbRecord, GWEN_DB_FLAGS_OVERWRITE_VARS,
		      "seq", seq);
  GWEN_Buffer_Reset(mbuf);
  if (LC_Card_CreateRecord(_card, 1, mbuf, dbRecord)) {
    DBG_ERROR(0, "Error creating record");
    GWEN_DB_Group_free(dbRecord);
    GWEN_Buffer_free(mbuf);
    return 0;
  }
  GWEN_Buffer_Rewind(mbuf);
  res=LC_Card_IsoUpdateRecord(_card, LC_CARD_ISO_FLAGS_RECSEL_GIVEN, 1,
                              GWEN_Buffer_GetStart(mbuf),
                              GWEN_Buffer_GetUsedBytes(mbuf));
  GWEN_DB_Group_free(dbRecord);
  GWEN_Buffer_free(mbuf);
  if (res!=LC_Client_ResultOk) {
    DBG_INFO(0, "here");
    return 0;
  }

  return seq;
}


int MediumDDV2::signKeyNumber() const {
  return _signKeyNumber;
}


int MediumDDV2::signKeyVersion() const {
  return _signKeyVersion;
}


int MediumDDV2::cryptKeyNumber() const {
  return _cryptKeyNumber;
}


int MediumDDV2::cryptKeyVersion() const {
  return _cryptKeyVersion;
}


Error MediumDDV2::selectContext(int country,
                                const string &instcode,
                                const string &userid){
  unsigned int i;

  if (_country==country &&
      _instcode==instcode &&
      _userid==userid) {
    return Error();
  }

  for (i=1; i<6; i++) {
    int lcountrycode;
    string linstcode;
    string luserid;
    string lserver;
    Error err;

    err=getContext(i, lcountrycode, linstcode, luserid, lserver);
    if (err.isOk()) {
      if (lcountrycode==country &&
          linstcode==instcode &&
          luserid==userid) {
        _userid=luserid;
        _instcode=linstcode;
        _country=lcountrycode;
        return Error();
      }
    }
  } /* for */

  return Error("MediumDDV2::selectContext",
	       ERROR_LEVEL_NORMAL,
	       HBCI_ERROR_CODE_INVALID,
	       ERROR_ADVISE_DONTKNOW,
               "no matching entry found");
}



Error MediumDDV2::getContext(int num,
                             int &countrycode,
                             string &instcode,
                             string &userid,
                             string &server) const {
  GWEN_DB_NODE *dbInst;
  LC_CLIENT_RESULT res;


  if (num<1)
    return Error("MediumDDV2::selectContext",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_INVALID,
		 ERROR_ADVISE_DONTKNOW,
		 "bad context number");
  if (!const_cast<MediumDDV2*>(this)->isMounted()){
    return Error("MediumDDV2::selectContext",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_INVALID,
		 ERROR_ADVISE_DONTKNOW,
		 "medium not mounted");
  }

  dbInst=GWEN_DB_Group_new("data");
  res=LC_DDVCard_ReadInstituteData(_card, num, dbInst);
  if (res!=LC_Client_ResultOk) {
    GWEN_DB_Group_free(dbInst);
    return Error("MediumDDV2::selectContext",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "context not available");
  }

  countrycode=GWEN_DB_GetIntValue(dbInst, "context/country", 0, 280);
  instcode=GWEN_DB_GetCharValue(dbInst, "context/bankCode", 0, "");
  userid=GWEN_DB_GetCharValue(dbInst, "context/userId", 0, "");
  server=GWEN_DB_GetCharValue(dbInst, "context/comAddress", 0, "");
  GWEN_DB_Group_free(dbInst);

  return Error();
}



Error MediumDDV2::changePIN() {
  return Error("MediumDDV2::changePIN",
	       "can't change PIN on DDV cards",
	       0);
}


Error MediumDDV2::changeContext(int context, int country,
			       const string instcode,
			       const string userid,
			       const string custid,
			       const string server) {
  Error err;

  if (context<1)
    return Error("MediumDDV2::changeContext",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_INVALID,
		 ERROR_ADVISE_DONTKNOW,
		 "bad context number");

  return Error("MediumDDV2::changeContext",
	       ERROR_LEVEL_NORMAL,
	       HBCI_ERROR_CODE_INVALID,
	       ERROR_ADVISE_DONTKNOW,
	       "not supported for now");

  err=mountMedium();
  if (!err.isOk())
    return Error("MediumDDV2::changeContext", err);


  return unmountMedium();
}



Error MediumDDV2::getProperty(const string &propertyName,
			     string &propertyValue) {
  if (strcasecmp(propertyName.c_str(), "loglevel")==0) {
    fprintf(stderr,
	    "WARNING: Property LOGLEVEL is no longer supported.\n"
	    "Please change your program accordingly, since in future "
	    "versions\n"
	    "this message will disappear\n");
  }
  else if (strcasecmp(propertyName.c_str(), "usekeypad")==0) {
    propertyValue=_useKeyPad?"yes":"no";
  }
  else {
    return Error("MediumDDV2::getProperty",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_UNKNOWN_PROPERTY,
		 ERROR_ADVISE_DONTKNOW,
		 "Unknown property",
		 propertyName);
  }

  return Error();
}


Error MediumDDV2::setProperty(const string &propertyName,
			     const string &propertyValue){
  if (strcasecmp(propertyName.c_str(), "loglevel")==0) {
    fprintf(stderr,
	    "WARNING: Property LOGLEVEL is no longer supported.\n"
	    "Please change your program accordingly, since in future "
	    "versions\n"
	    "this message will disappear\n");
  }
  else if (strcasecmp(propertyName.c_str(), "usekeypad")==0) {
    if (strcasecmp(propertyValue.c_str(),"yes")==0) {
      _useKeyPad=true;
    }
    else if (strcasecmp(propertyValue.c_str(),"true")==0) {
      _useKeyPad=true;
    }
    else if (strcasecmp(propertyValue.c_str(),"1")==0) {
      _useKeyPad=true;
    }
    else if (strcasecmp(propertyValue.c_str(),"no")==0) {
      _useKeyPad=false;
    }
    else if (strcasecmp(propertyValue.c_str(),"false")==0) {
      _useKeyPad=false;
    }
    else if (strcasecmp(propertyValue.c_str(),"0")==0) {
      _useKeyPad=false;
    }
    else {
      return Error("MediumDDV2::setProperty",
		   ERROR_LEVEL_NORMAL,
		   HBCI_ERROR_CODE_UNKNOWN_PROPERTY,
		   ERROR_ADVISE_DONTKNOW,
		   "Bad value for property \"usekeypad\"",
		   propertyValue);
    }
  }
  else {
    return Error("MediumDDV2::setProperty",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_UNKNOWN_PROPERTY,
		 ERROR_ADVISE_DONTKNOW,
		 "Unknown property",
		 propertyName);
  }

  return Error();
}


Error MediumDDV2::_readKeys() {
  _signKeyNumber=LC_DDVCard_GetSignKeyNumber(_card);
  _signKeyVersion=LC_DDVCard_GetSignKeyVersion(_card);
  _cryptKeyNumber=LC_DDVCard_GetCryptKeyNumber(_card);
  _cryptKeyVersion=LC_DDVCard_GetCryptKeyVersion(_card);
  if (_signKeyNumber!=-1 && _signKeyVersion!=-1 &&
      _cryptKeyNumber!=-1 && _cryptKeyVersion!=-1)
    return Error();
  return Error("MediumDDV2::_readKeys",
	       ERROR_LEVEL_NORMAL,
	       HBCI_ERROR_CODE_INVALID,
	       ERROR_ADVISE_DONTKNOW,
	       "error reading keys");
}



} // namespace HBCI

HBCI_Medium *HBCI_MediumDDV2_Medium(HBCI_MediumDDV2 *h){
  return h;
}

HBCI_MediumDDV2 *HBCI_Medium_MediumDDV2(HBCI_Medium *h){
  try {
    return dynamic_cast<HBCI_MediumDDV2*>(h);
  }
  catch (bad_cast) {
    fprintf(stderr,
	    "HBCI_Medium_MediumDDV2: Caught bad_cast exception.\n");
    return 0l;
  }
}



